package amphis;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import graph.Undirected;

/*
   L'Âne rouge
   Cf https://fr.wikipedia.org/wiki/L%27%C3%82ne_rouge
   
   On représente la grille comme ceci :

       j 0 1 2 3
     i  +-+-+-+-+
      0 | | | | |
        +-+-+-+-+
      1 | | | | |
        +-+-+-+-+
      2 | | | | |
        +-+-+-+-+
      3 | | | | |
        +-+-+-+-+
      4 | | | | |
        +-+-+-+-+
 */


// Une pièce est définie par sa hauteur `h`, sa largeur `w` et sa position.
class Block implements Comparable<Block> {
  final int h, w;
  final int i, j;
  
  Block(int h, int w, int i, int j) {
    super();
    this.h = h;
    this.w = w;
    this.i = i;
    this.j = j;
  }

  char color() {
    if (this.h == 1)
      return this.w == 1 ? 'A' : 'B';
    else
      return this.w == 1 ? 'C' : 'D';
  }
  
  @Override
  public String toString() {
    return "(" + h + "x" + w + " " + i + "," + j + ")";
  }
  
  // ordre lexicographique
  @Override
  public int compareTo(Block that) {
    if (this.h != that.h) return this.h - that.h;
    if (this.w != that.w) return this.w - that.w;
    if (this.i != that.i) return this.i - that.i;
    return this.j - that.j;
  }
  
  @Override
  public boolean equals(Object o) {
    if (!(o instanceof Block)) return false;
    Block that = (Block)o;
    return this.compareTo(that) == 0;
  }
  
  @Override
  public int hashCode() {
    return this.h + 31 * (this.w + 31 * (this.i + 31 * this.j));
  }
  
}

// Une configuration est une liste *triée* de pièces
// --- en triant, on garantit l'unicité.
class State {
  final LinkedList<Block> blocks; // /!\ triée
  
  State(List<Block> l) {
    this();
    for (Block b: l) // /!\ on fait une copie de la liste l
      this.blocks.add(b);
    this.blocks.sort(null);
  }
  
  State() {
    this.blocks = new LinkedList<>();
  }
  
  // la configuration initiale du jeu de l'âne rouge
  final static State start = new State();
  static {
    // A 1x1
    List<Block> l = start.blocks;
    l.add(new Block(1, 1, 3, 1));
    l.add(new Block(1, 1, 3, 2));
    l.add(new Block(1, 1, 4, 0));
    l.add(new Block(1, 1, 4, 3));
    // B 1x2
    l.add(new Block(1, 2, 2, 1));
    // C 2x1
    l.add(new Block(2, 1, 0, 0));
    l.add(new Block(2, 1, 0, 3));
    l.add(new Block(2, 1, 2, 0));
    l.add(new Block(2, 1, 2, 3));
    // D 2x2
    l.add(new Block(2, 2, 0, 1));
  }
  
  boolean success() {
    for (Block b: this.blocks)
      if (b.h == 2 && b.w == 2 && b.i == 3 && b.j == 1)
        return true;
    return false;
  }
    
  public String toString() {
    char[][] m = new char[5][4];
    for (int i = 0; i < 5; i++)
      for (int j = 0; j < 4; j++)
        m[i][j] = ' ';
    for (Block b: this.blocks) {
      char c = b.color();
      for (int i = 0; i < b.h; i++)
        for (int j = 0; j < b.w; j++)
          m[b.i + i][b.j + j] = c;
    }
    StringBuffer b = new StringBuffer();
    for (int i = 0; i < 5; i++) {
      for (int j = 0; j < 4; j++)
        b.append(m[i][j]);
      b.append('\n');
    }
    return b.toString();
  }
  
  @Override
  public int hashCode() {
    int h = 0;
    for (Block b: this.blocks)
      h = 31 * h + b.hashCode();
    return h;
  }
  
  @Override
  public boolean equals(Object o) {
    if (!(o instanceof State)) return false;
    State that = (State)o;
    Iterator<Block> it1 = this.blocks.iterator();
    Iterator<Block> it2 = that.blocks.iterator();
    while (it1.hasNext()) {
      assert(it2.hasNext());
      Block b1 = it1.next(), b2 = it2.next();
      if (!b1.equals(b2)) return false;
    }
    assert(!it2.hasNext());
    return true;
  }
  
  // déplacements
  List<State> moves() {
    boolean[][] free = new boolean[5][4];
    for (int i = 0; i < 5; i++)
      for (int j = 0; j < 4; j++)
        free[i][j] = true;
    for (Block b: this.blocks)
      for (int i = 0; i < b.h; i++)
        for (int j = 0; j < b.w; j++)
          free[b.i + i][b.j + j] = false;
    Block[] a = this.blocks.toArray(new Block[0]);
    LinkedList<State> ml = new LinkedList<>();
    for (int k = 0; k < a.length; k++) {
      Block b = a[k];
      int i = b.i, j = b.j, h = b.h, w = b.w;
      // vers le haut ?
      if (i > 0 && free[i-1][j] && (w == 1 || free[i-1][j+1])) {
        ml.add(createMove(a, k, new Block(h, w, i-1, j)));
      }
      // vers le bas ?
      if (i+h < 5 && free[i+h][j] && (w == 1 || free[i+h][j+1])) {
        ml.add(createMove(a, k, new Block(h, w, i+1, j)));
      }
      // vers la gauche ?
      if (j > 0 && free[i][j-1] && (h == 1 || free[i+1][j-1])) {
        ml.add(createMove(a, k, new Block(h, w, i, j-1)));
      }
      // vers la droite ?
      if (j+w < 4 && free[i][j+w] && (h == 1 || free[i+1][j+w])) {
        ml.add(createMove(a, k, new Block(h, w, i, j+1)));
      }
      a[k] = b;
    }
    return ml;
  }
    
  private State createMove(Block[] a, int k, Block b) {
    a[k] = b;
    LinkedList<Block> l = new LinkedList<>();
    for (Block x: a) l.add(x);
    return new State(l);
  }
  
}

public class AneRouge {
  
  public static void main(String[] args) {
    System.out.println(State.start);
    
    // recherche d'une solution de longueur minimale
    double start = System.currentTimeMillis();
    bfs(State.start);
    System.err.println("  " + (System.currentTimeMillis() - start) + " ms");
    start = System.currentTimeMillis();
    
    // construction du graphe complet
    Undirected<State> g = buildGraph();
    System.out.println(g.vertices().size() + " sommets");
    System.out.println(g.edges().size() + " arcs");
    List<List<State>> comp = g.components();
    System.out.println(comp.size() + " composantes connexes");
  }
  
  // comme dans graph.BFSPath, mais en s'arrêtant dès qu'une solution est atteinte
  static boolean bfs(State source) {
    System.out.println("Parcours en largeur :");
    HashMap<State, Integer> dist = new HashMap<>();
    HashMap<State, State> path = new HashMap<>();
    Queue<State> q = new LinkedList<>();
    q.add(source);
    path.put(source, source);
    dist.put(source,  0);
    while (!q.isEmpty()) {
      State v = q.poll();
      int d = dist.get(v);
      if (v.success()) {
        System.out.println("  " + path.size() + " états explorés");
        System.out.println("  solution de longueur " + d);
        // printPath(path, source, v);
        return true;
      }
      for (State s: v.moves())
        if (!dist.containsKey(s)) {
          dist.put(s, d+1);
          path.put(s, v);
          q.add(s);
        }
    }
    return false;
  }
  
  static void printPath(Map<State, State> path, State source, State target) {
    if (!source.equals(target)) {
      printPath(path, source, path.get(target));
    }
    System.out.println(target);
  }
  
  static int[][] blocks = {
    {2,2}, {2,1}, {2,1}, {2,1}, {2,1}, {1,2}, {1,1}, {1,1}, {1,1}, {1,1}
  };

  static Undirected<State> buildGraph() {
    Undirected<State> g = new Undirected<>();
    LinkedList<Block> l = new LinkedList<>();
    boolean[][] free = new boolean[5][4];
    for (int i = 0; i < 5; i++)
      for (int j = 0; j < 4; j++)
        free[i][j] = true;
    addBlocks(g, l, free, 0);
    for (State v: g.vertices())
      for (State w: v.moves())
        g.addEdge(v, w);
    return g;
  }
  private static void addBlocks(Undirected<State> g, LinkedList<Block> l, boolean[][] free, int k) {
    if (k == blocks.length) {
      g.addVertex(new State(l));
      return;
    }
    int h = blocks[k][0], w = blocks[k][1];
    for (int i = 0; i <= 5 - h; i++)
      for (int j = 0; j <= 4 - w; j++)
        if (free[i][j] && (w == 1 || free[i][j+1])
                       && (h == 1 || free[i+1][j])
                       && (w == 1 || h == 1 || free[i+1][j+1])) {
          Block b = new Block(h, w, i, j);
          if (l.isEmpty() || b.compareTo(l.getFirst()) < 0) {
            for (int di = 0; di < h; di++)
              for (int dj = 0; dj < w; dj++)
                free[i+di][j+dj] = false;
            l.addFirst(b);  
            addBlocks(g, l, free, k+1);
            l.removeFirst();
            for (int di = 0; di < h; di++)
              for (int dj = 0; dj < w; dj++)
                free[i+di][j+dj] = true;
          }
        }
  }
  
}
